/*****************************************************************************

  Licensed to Accellera Systems Initiative Inc. (Accellera) under one or
  more contributor license agreements.  See the NOTICE file distributed
  with this work for additional information regarding copyright ownership.
  Accellera licenses this file to you under the Apache License, Version 2.0
  (the "License"); you may not use this file except in compliance with the
  License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
  implied.  See the License for the specific language governing
  permissions and limitations under the License.

 *****************************************************************************/

#include <systemc-ams>
#include "test_utilities.h"

test_util::sorted_stream sstr;

SCA_TDF_MODULE(delay)
{
  sca_tdf::sca_in<double>  inp;
  sca_tdf::sca_out<double> outp;

  void set_attributes();
  void initialize();
  void processing();

  SCA_CTOR(delay) : inp("inp"), outp("outp") {}
};

void delay::set_attributes()
{
  outp.set_delay(2);
}

void delay::initialize()
{
  outp.initialize(0.77777);
  outp.initialize(0.333, 1);
}

void delay::processing()
{
  outp = inp;
}

/////////////////////////////

SCA_TDF_MODULE(adder)
{
  sca_tdf::sca_in<double>  inp1;
  sca_tdf::sca_in<double>  inp2;
  sca_tdf::sca_out<double> outp;

  void processing();

  SCA_CTOR(adder) : inp1("inp1"), inp2("inp2"), outp("outp") {}
};

void adder::processing()
{
  outp.write(inp1 + inp2.read());
}

/////////////////////////////

SCA_TDF_MODULE(dec)
{
  sca_tdf::sca_in<double> inp;
  sca_tdf::sca_out<double> outp;

  void set_attributes()
  {
    inp.set_rate(2);
  }

  void processing();

  SCA_CTOR(dec) : inp("inp"), outp("outp") {}
};

void dec::processing()
{
  outp.write((inp.read(0) + inp[1]) / 2.0);
}

/////////////////////////////

SCA_TDF_MODULE(intp)
{
  sca_tdf::sca_in<double> inp;
  sca_tdf::sca_out<double> outp;

  void set_attributes()
  {
    outp.set_rate(2);
  }

  void processing();

  SCA_CTOR(intp)
  : inp("inp"), outp("outp"), last_value(0.0)
  {}

 private:
  double last_value;
};

void intp::processing()
{
  double diff2 = (inp - last_value) / 2.0;
  outp[0] = last_value + diff2;
  outp.write(inp, 1);

  last_value = inp.read();
}

///////////////////////////////////

SCA_TDF_MODULE(const_src)
{
  sca_tdf::sca_out<double> outp;

  void set_attributes();
  void processing();

  SCA_CTOR(const_src) : outp("outp") {}
};

void const_src::set_attributes()
{
  outp.set_timestep(1.0, sc_core::SC_MS);
}

void const_src::processing()
{
  outp = 1.0;
}

SCA_TDF_MODULE(sink)
{
  sca_tdf::sca_in<double> inp;

  void processing();

  SCA_CTOR(sink) : inp("inp") {}
};

void sink::processing()
{
  std::ostringstream str;
  str << get_time() << "  " << name()
      << " current value: " << inp.read();

  sstr.redirect_cout(get_time());

  SC_REPORT_INFO("Simple_TDF", str.str().c_str());

  sstr.restore_cout();
}

////////////////////////////////

int sc_main(int argc, char* argv[])
{
  TEST_LABEL_START;

  sca_tdf::sca_signal<double> sig1("sig1");
  sca_tdf::sca_signal<double> sig2;
  sca_tdf::sca_signal<double> sig3("sig3");
  sca_tdf::sca_signal<double> sig4("sig4");
  sca_tdf::sca_signal<double> sig5("sig5");

  adder add1("add1");
  add1.inp1(sig1);
  add1.inp2(sig2);
  add1.outp(sig3);

  dec dec1("dec1");
  dec1.inp(sig3);
  dec1.outp(sig4);

  intp intp1("intp1");
  intp1.inp(sig4);
  intp1.outp(sig5);

  delay del1("del1");
  del1(sig5, sig2); // positional binding

  const_src src1("src1");
  src1.outp(sig1);

  sink sinkdec("sinkdec");
  sinkdec.inp(sig4);

  sink sinkdel("sinkdel");
  sinkdel.inp(sig2);

  sc_core::sc_start(10.0, sc_core::SC_MS);

  sstr.print();

  TEST_LABEL_END;

  return 0;
}
